import importlib
import json
from pathlib import Path
from os import path
import numpy as np
import os
import uuid
import time
import shutil

from gfootball.env import football_action_set


def renderer(state, env):
    html_renderer(env)


def run_right_agent(obs):
    # keep running right.
    return [5] * obs.controlled_players


def run_left_agent(obs):
    # keep running left.
    return [1] * obs.controlled_players


def do_nothing_agent(obs):
    # do nothing.
    return [0] * obs.controlled_players


def builtin_ai_agent(obs):
    # execute builtin AI behavior.
    return [19] * obs.controlled_players


agents = {
    "run_right": run_right_agent,
    "run_left": run_left_agent,
    "do_nothing": do_nothing_agent,
    "builtin_ai": builtin_ai_agent,
}


def parse_single_player(obs_raw_entry):
    # Remove pixel information.
    if "frame" in obs_raw_entry:
        del obs_raw_entry["frame"]
    for k, v in obs_raw_entry.items():
        if type(v) == np.ndarray:
            obs_raw_entry[k] = v.tolist()
    return obs_raw_entry


def try_get_video(env, keep_running=False):
    if not env.football_video_path:
        internal_env = m_envs[env.configuration.id]
        while not hasattr(internal_env, '_env'):
            internal_env = internal_env.env
        if not keep_running and internal_env._env._step == -1:
            # Generate no-op step, so that video is available.
            internal_env.step([0] * (env.configuration.team_1 + env.configuration.team_2))
        trace = internal_env._env._trace
        if trace:
          trace._dump_config['episode_done']._min_frequency = 0
          dumps = trace.process_pending_dumps(True)
          env.football_video_path = retrieve_video_link(dumps)
        if not env.football_video_path:
            return
        if keep_running:
            trace.write_dump('episode_done')
    if 'LiveVideoPath' in env.info and env.info['LiveVideoPath'] is not None:
        target_path = Path(env.info['LiveVideoPath'])
        target_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(env.football_video_path, target_path)
        env.football_video_path = env.info['LiveVideoPath']


def update_observations_and_rewards(configuration, state, obs, rew=None):
    """Updates agent-visible observations given 'raw' observations from environment.
    Observations in 'obs' are coming directly from the environment and are in 'raw' format.
    """
    state[0].observation.controlled_players = configuration.team_1
    state[1].observation.controlled_players = configuration.team_2

    assert len(obs) == configuration.team_1 + configuration.team_2
    if rew is not None:
        state[0].reward = rew
        state[1].reward = -rew
    state[0].observation.players_raw = [
        parse_single_player(obs[x]) for x in range(configuration.team_1)
    ]
    state[1].observation.players_raw = [
        parse_single_player(obs[x + configuration.team_1])
        for x in range(configuration.team_2)
    ]


def mark_invalid(agent, message):
    agent.status = "INVALID"
    agent.reward = -100
    agent.info.debug_info = message


def maybe_terminate(env, state):
    if state[0].status != "ACTIVE" or state[1].status != "ACTIVE":
        if state[0].status == "ACTIVE":
            state[0].status = "DONE"
            state[0].reward = 100
            state[0].info.debug_info = "Opponent forfeited. You win."
        elif not state[0].reward:
            state[0].reward = -100
        if state[1].status == "ACTIVE":
            state[1].status = "DONE"
            state[1].reward = 100
            state[1].info.debug_info = "Opponent forfeited. You win."
        elif not state[1].reward:
            state[1].reward = -100
        try_get_video(env)
        return True
    return False


def football_env():
    # Use lazy-import to avoid this heavy dependency unless it is really needed.
    return importlib.import_module("gfootball.env")


# Global dictionary with active environments.
m_envs = {}


def cleanup(env):
    global m_envs
    del m_envs[env.configuration.id]


def cleanup_all():
    global m_envs
    del m_envs


def retrieve_video_link(dumps):
    for entry in dumps:
        if entry['name'] == 'episode_done':
            print("Received video link.")
            return entry['video']
    return None


def interpreter(state, env):
    global m_envs
    if "id" not in env.configuration or env.configuration.id is None:
        env.configuration.id = str(uuid.uuid4())

    if (env.configuration.id not in m_envs) or env.done:
        if env.configuration.id not in m_envs:
            print("Staring a new environment %s: with scenario: %s" %
                  (env.configuration.id, env.configuration.scenario_name))

            other_config_options = {}
            # Use webm to encode videos (so that you can see them in the browser).
            other_config_options["video_format"] = "webm"
            if env.configuration.running_in_notebook:
                assert not env.configuration.render, "Render is not supported inside notebook environment."

            env.football_video_path = None
            if 'TeamNames' in env.info:
                names = env.info['TeamNames']
                assert len(names) == 2
                other_config_options['custom_display_stats'] = [
                    'LEFT PLAYER: %s' % names[0],
                    'RIGHT PLAYER: %s' % names[1]]
            m_envs[env.configuration.id] = football_env().create_environment(
                env_name=env.configuration.scenario_name,
                stacked=False,
                # We use 'raw' representation to transfer data between server and agents.
                representation='raw',
                logdir=path.join(env.configuration.logdir, env.configuration.id),
                write_goal_dumps=False,
                write_full_episode_dumps=env.configuration.save_video,
                write_video=env.configuration.save_video,
                render=env.configuration.render,
                number_of_left_players_agent_controls=env.configuration.team_1,
                number_of_right_players_agent_controls=env.configuration.team_2,
                other_config_options=other_config_options)
        else:
            print("Resetting environment %s: with scenario: %s" %
                  (env.configuration.id, env.configuration.scenario_name))
        obs = m_envs[env.configuration.id].reset()
        update_observations_and_rewards(configuration=env.configuration,
                                        state=state,
                                        obs=obs)
    if env.done:
        return state

    if maybe_terminate(env, state):
        return state

    # verify actions.
    controlled_players = env.configuration.team_1
    action_set = football_action_set.action_set_dict['default']

    try:
        for action in state[0].action:
            football_action_set.named_action_from_action_set(action_set, action)
    except Exception:
        mark_invalid(state[0], "Invalid action provided: %s." % state[0].action)
    if len(state[0].action) != env.configuration.team_1:
        mark_invalid(state[0], "Invalid number of actions provided: Expected %d, got %d." %
            (env.configuration.team_1, len(state[0].action)))
    actions_to_env = state[0].action

    try:
        for action in state[1].action:
            football_action_set.named_action_from_action_set(action_set, action)
    except Exception:
        mark_invalid(state[1], "Invalid action provided: %s." % state[1].action)
    if len(state[1].action) != env.configuration.team_2:
        mark_invalid(state[1], "Invalid number of actions provided: Expected %d, got %d." %
            (env.configuration.team_2, len(state[1].action)))
    if env.configuration.team_2:
        actions_to_env = actions_to_env + state[1].action

    if maybe_terminate(env, state):
        return state
    obs, rew, done, info = m_envs[env.configuration.id].step(actions_to_env)

    if "dumps" in info:
        env.football_video_path = retrieve_video_link(info["dumps"])
    update_observations_and_rewards(configuration=env.configuration,
                                    state=state,
                                    obs=obs,
                                    rew=obs[0]['score'][0]-obs[0]['score'][1])

    ## TODO: pass other information from 'info' to the state/agent.
    if done:
        for agent in range(2):
            state[agent].status = "DONE"
        try_get_video(env)

    return state


dirpath = path.dirname(__file__)
jsonpath = path.abspath(path.join(dirpath, "football.json"))
with open(jsonpath) as f:
    specification = json.load(f)


def html_renderer(env):
    try_get_video(env, keep_running=True)
    if not env.football_video_path:
        raise Exception(
            "No video found. Was environment created with save_video enabled?"
        )

    from IPython.display import display, HTML
    from base64 import b64encode

    video = open(env.football_video_path, 'rb').read()
    env.football_video_path = None
    data_url = "data:video/webm;base64," + b64encode(video).decode()

    html = """
<video width=800 controls>
  <source src="%s" type="video/webm">
</video>
""" % data_url
    display(HTML(html))
    return ""



# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from os import path
from random import choice

checks = [
    [0, 1, 2],
    [3, 4, 5],
    [6, 7, 8],
    [0, 3, 6],
    [1, 4, 7],
    [2, 5, 8],
    [0, 4, 8],
    [2, 4, 6],
]

EMPTY = 0


def random_agent(obs):
    return choice([c for c in range(len(obs.board)) if obs.board[c] == EMPTY])


def reaction_agent(obs):
    # Connect 3 in a row to win.
    for check in checks:
        left = list(filter(lambda c: obs.board[c] != obs.mark, check))
        if len(left) == 1 and obs.board[left[0]] == EMPTY:
            return left[0]

    # Block 3 in a row to prevent loss.
    opponent = 2 if obs.mark == 1 else 1
    for check in checks:
        left = list(filter(lambda c: obs.board[c] != opponent, check))
        if len(left) == 1 and obs.board[left[0]] == EMPTY:
            return left[0]

    # No 3-in-a-rows, return random unmarked.
    return choice(list(filter(lambda m: m[1] == EMPTY, enumerate(obs.board))))[0]


agents = {"random": random_agent, "reaction": reaction_agent}


def interpreter(state, env):
    if env.done:
        return state

    # Isolate the active and inactive agents.
    active = state[0] if state[0].status == "ACTIVE" else state[1]
    inactive = state[0] if state[0].status == "INACTIVE" else state[1]
    if active.status != "ACTIVE" or inactive.status != "INACTIVE":
        active.status = "DONE" if active.status == "ACTIVE" else active.status
        inactive.status = "DONE" if inactive.status == "INACTIVE" else inactive.status
        return state

    # The board is shared, only update the first state.
    board = state[0].observation.board

    # Illegal move by the active agent.
    if board[active.action] != EMPTY:
        active.status = f"Invalid move: {active.action}"
        inactive.status = "DONE"
        return state

    # Mark the position.
    board[active.action] = active.observation.mark

    # Check for a win.
    if any(all(board[p] == active.observation.mark for p in c) for c in checks):
        active.reward = 1
        active.status = "DONE"
        inactive.reward = -1
        inactive.status = "DONE"
        return state

    # Check for a tie.
    if all(mark != EMPTY for mark in board):
        active.status = "DONE"
        inactive.status = "DONE"
        return state

    # Swap active and inactive agents to switch turns.
    active.status = "INACTIVE"
    inactive.status = "ACTIVE"

    return state


def renderer(state, env):
    row_bar = "\n---+---+---\n"
    marks = [" ", "X", "O"]

    def print_pos(pos):
        str = ""
        if pos % 3 == 0 and pos > 0:
            str += row_bar
        if pos % 3 != 0:
            str += "|"
        return str + f" {marks[state[0].observation.board[pos]]} "

    return "".join(print_pos(p) for p in range(9))


jsonpath = path.abspath(path.join(path.dirname(__file__), "tictactoe.json"))
with open(jsonpath) as f:
    specification = json.load(f)


def html_renderer():
    jspath = path.abspath(path.join(path.dirname(__file__), "tictactoe.js"))
    with open(jspath) as f:
        return f.read()



from kaggle_environments import make
from .agents import rock, paper, agents


def negative_move_agent(observation, configuration):
    return -1


def too_big_sign_agent(observation, configuration):
    return 1000000


def non_integer_agent(observation, configuration):
    return 0.3


def none_agent(observation, configuration):
    return None


def test_rps_completes():
    env = make("rps", configuration={"episodeSteps": 10, "tieRewardThreshold": 1})
    env.run([rock, rock])
    json = env.toJSON()
    assert json["name"] == "rps"
    assert json["statuses"] == ["DONE", "DONE"]


def test_all_agents():
    env = make("rps", configuration={"episodeSteps": 3, "tieRewardThreshold": 1})
    for agent in agents:
        env.run([agent, agent])
        json = env.toJSON()
        assert json["statuses"] == ["DONE", "DONE"]


def test_tie():
    env = make("rps", configuration={"episodeSteps": 3, "tieRewardThreshold": 1})
    env.run([rock, rock])
    assert env.render(mode='ansi') == "Round 1: Rock vs Rock, Score: 0 to 0\nRound 2: Rock vs Rock, Score: 0 to 0\nGame ended on round 2, final score: 0 to 0\n"
    json = env.toJSON()
    assert json["rewards"] == [0, 0]
    assert json["statuses"] == ["DONE", "DONE"]


def test_threshold_tie():
    env = make("rps", configuration={"episodeSteps": 3, "tieRewardThreshold": 4})
    env.run([rock, paper])
    assert env.render(mode='ansi') == "Round 1: Rock vs Paper, Score: -1.0 to 1.0\nRound 2: Rock vs Paper, Score: 0 to 0\nGame ended on round 2, final score: 0 to 0\n"
    json = env.toJSON()
    assert json["rewards"] == [0, 0]
    assert json["statuses"] == ["DONE", "DONE"]


def test_win():
    env = make("rps", configuration={"episodeSteps": 2, "tieRewardThreshold": 1})
    env.run([paper, rock])
    json = env.toJSON()
    print(json)
    assert json["rewards"] == [1, -1]
    assert json["statuses"] == ["DONE", "DONE"]


def test_loss():
    env = make("rps", configuration={"episodeSteps": 2, "tieRewardThreshold": 1})
    env.run([rock, paper])
    json = env.toJSON()
    assert json["rewards"] == [-1, 1]
    assert json["statuses"] == ["DONE", "DONE"]


def test_negative_move():
    env = make("rps", configuration={"episodeSteps": 10, "tieRewardThreshold": 1})
    env.run([negative_move_agent, rock])
    json = env.toJSON()
    assert json["rewards"] == [None, 1]
    assert json["statuses"] == ['INVALID', 'DONE']


def test_non_integer_move():
    env = make("rps", configuration={"episodeSteps": 10, "tieRewardThreshold": 1})
    env.run([non_integer_agent, rock])
    json = env.toJSON()
    assert json["rewards"] == [None, 1]
    assert json["statuses"] == ['INVALID', 'DONE']


def test_too_big_move():
    env = make("rps", configuration={"episodeSteps": 10, "tieRewardThreshold": 1})
    env.run([paper, too_big_sign_agent])
    json = env.toJSON()
    assert json["rewards"] == [1, None]
    assert json["statuses"] == ['DONE', 'INVALID']

def test_agent_reward():
    env = make("rps", configuration={"episodeSteps": 2, "tieRewardThreshold": 1})
    env.run([paper, rock])
    json = env.toJSON()
    last_step = json["steps"][-1]
    assert last_step[0]["observation"]["step"] == last_step[0]["observation"]["reward"]
    assert last_step[0]["observation"]["reward"] == -last_step[1]["observation"]["reward"]
    assert json["statuses"] == ["DONE", "DONE"]



from .constants import Constants
from .game_map import GameMap
from .game_objects import Player, Unit, City, CityTile

INPUT_CONSTANTS = Constants.INPUT_CONSTANTS


class Game:
    def _initialize(self, messages):
        """
        initialize state
        """
        self.id = int(messages[0])
        self.turn = -1
        # get some other necessary initial input
        mapInfo = messages[1].split(" ")
        self.map_width = int(mapInfo[0])
        self.map_height = int(mapInfo[1])
        self.map = GameMap(self.map_width, self.map_height)
        self.players = [Player(0), Player(1)]

    def _end_turn(self):
        print("D_FINISH")

    def _reset_player_states(self):
        self.players[0].units = []
        self.players[0].cities = {}
        self.players[0].city_tile_count = 0
        self.players[1].units = []
        self.players[1].cities = {}
        self.players[1].city_tile_count = 0

    def _update(self, messages):
        """
        update state
        """
        self.map = GameMap(self.map_width, self.map_height)
        self.turn += 1
        self._reset_player_states()

        for update in messages:
            if update == "D_DONE":
                break
            strs = update.split(" ")
            input_identifier = strs[0]
            if input_identifier == INPUT_CONSTANTS.RESEARCH_POINTS:
                team = int(strs[1])
                self.players[team].research_points = int(strs[2])
            elif input_identifier == INPUT_CONSTANTS.RESOURCES:
                r_type = strs[1]
                x = int(strs[2])
                y = int(strs[3])
                amt = int(float(strs[4]))
                self.map._setResource(r_type, x, y, amt)
            elif input_identifier == INPUT_CONSTANTS.UNITS:
                unittype = int(strs[1])
                team = int(strs[2])
                unitid = strs[3]
                x = int(strs[4])
                y = int(strs[5])
                cooldown = float(strs[6])
                wood = int(strs[7])
                coal = int(strs[8])
                uranium = int(strs[9])
                self.players[team].units.append(Unit(team, unittype, unitid, x, y, cooldown, wood, coal, uranium))
            elif input_identifier == INPUT_CONSTANTS.CITY:
                team = int(strs[1])
                cityid = strs[2]
                fuel = float(strs[3])
                lightupkeep = float(strs[4])
                self.players[team].cities[cityid] = City(team, cityid, fuel, lightupkeep)
            elif input_identifier == INPUT_CONSTANTS.CITY_TILES:
                team = int(strs[1])
                cityid = strs[2]
                x = int(strs[3])
                y = int(strs[4])
                cooldown = float(strs[5])
                city = self.players[team].cities[cityid]
                citytile = city._add_city_tile(x, y, cooldown)
                self.map.get_cell(x, y).citytile = citytile
                self.players[team].city_tile_count += 1;
            elif input_identifier == INPUT_CONSTANTS.ROADS:
                x = int(strs[1])
                y = int(strs[2])
                road = float(strs[3])
                self.map.get_cell(x, y).road = road



from kaggle_environments.envs.lux_ai_2021.test_agents.python.random_agent import random_agent
import random
from .test_agents.js_simple.main import js_agent as js_simple_agent
from .test_agents.python.random_agent import random_agent
from .test_agents.python.simple_agent import agent as simple_agent
agents = {
    "random_agent": random_agent,
    "simple_agent": simple_agent,
    "js_simple_agent": js_simple_agent,
}


import json
from os import path
from kaggle_environments.helpers import *


class Observation(Observation):
    """
    Observation primarily used as a helper to construct the State from the raw observation.
    This provides bindings for the observation type described at https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/rps/rps.json
    """
    @property
    def last_opponent_action(self) -> int:
        """Move the opponent took on the last turn."""
        return self["halite"]


class Configuration(Configuration):
    """
    Configuration provides access to tunable parameters in the environment.
    This provides bindings for the configuration type described at https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/rps/rps.json
    """
    @property
    def signs(self) -> int:
        """Number of choices each step (3 for the normal rock, paper, scissors)"""
        return self["signs"]



# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from kaggle_environments import make, evaluate

env = None


def before_each(state=None, configuration=None):
    global env
    steps = [] if state == None else [state]
    env = make("connectx", steps=steps,
               configuration=configuration, debug=True)


def test_has_correct_timeouts():
    before_each()
    assert env.configuration.actTimeout == 2


def test_can_train_first():
    before_each()
    trainer = env.train([None, 'random'])
    obs = trainer.reset()
    assert "board" in obs
    obs, _, _, _ = trainer.step(0)
    assert "board" in obs


def test_can_train_second():
    before_each()
    trainer = env.train(['random', None])
    obs = trainer.reset()
    assert "board" in obs
    obs, _, _, _ = trainer.step(0)
    assert "board" in obs


def test_to_json():
    before_each()
    json = env.toJSON()
    assert json["name"] == "connectx"
    assert json["rewards"] == [0, 0]
    assert json["statuses"] == ["ACTIVE", "INACTIVE"]
    assert json["specification"]["reward"]["type"] == ["number", "null"]


def test_can_reset():
    before_each()
    assert env.reset() == [
        {
            "action": 0,
            "status": "ACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "board": [0] * 42, "mark": 1, "step": 0},
            "reward": 0,
        },
        {
            "action": 0,
            "status": "INACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_mark():
    before_each(configuration={"rows": 4, "columns": 5, "inarow": 3})
    assert env.step([2, None]) == [
        {
            "action": 2,
            "status": "INACTIVE",
            "info": {},
            "observation": {
                "remainingOverageTime": 60,
                "board": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                "mark": 1,
                "step": 1,
            },
            "reward": 0,
        },
        {
            "action": 0,
            "status": "ACTIVE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_mark_out_of_bounds():
    before_each(configuration={"rows": 4, "columns": 5, "inarow": 3})
    assert env.step([10, None]) == [
        {
            "action": 10,
            "status": "INVALID",
            "info": {},
            "observation": {
                "remainingOverageTime": 60,
                "board": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                "mark": 1,
                "step": 1,
            },
            "reward": None,
        },
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_mark_a_full_column():
    board = [1, 2, 0, 0, 0, 1, 2, 0, 0, 0, 1, 2, 0, 0, 0, 1, 2, 0, 0, 0]
    before_each(
        configuration={"rows": 4, "columns": 5, "inarow": 3},
        state=[{"observation": {"board": board}},
               {"observation": {}}],
    )
    assert env.step([1, None]) == [
        {
            "action": 1,
            "status": "INVALID",
            "info": {},
            "observation": {"remainingOverageTime": 60, "board": board, "mark": 1, "step": 1},
            "reward": None,
        },
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_win():
    board = [0, 0, 0, 0, 0, 1, 2, 0, 0, 0, 1, 2, 0, 0, 0, 1, 2, 0, 0, 0]
    board_post_move = board[:]
    board_post_move[0] = 1
    before_each(
        configuration={"rows": 4, "columns": 5, "inarow": 3},
        state=[{"observation": {"board": board}},
               {"observation": {}}],
    )
    assert env.step([0, None]) == [
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "board": board_post_move, "mark": 1, "step": 1},
            "reward": 1,
        },
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": -1,
        },
    ]


def test_can_tie():
    board = [0, 0, 1, 2, 1, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1]
    board_post_move = board[:]
    board_post_move[0] = 1
    board_post_move[1] = 2
    before_each(
        configuration={"rows": 4, "columns": 5, "inarow": 3},
        state=[{"observation": {"remainingOverageTime": 60, "board": board}}, {"observation": {}}],
    )
    env.step([0, None])
    assert env.step([None, 1]) == [
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "board": board_post_move, "mark": 1, "step": 2},
            "reward": 0,
        },
        {
            "action": 1,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": 0,
        },
    ]


def test_can_render():
    board = [0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 1, 2, 1, 0, 1, 2, 1, 2, 1]
    before_each(
        configuration={"rows": 4, "columns": 5, "inarow": 3},
        state=[{"observation": {"remainingOverageTime": 60, "board": board}}, {"observation": {}}],
    )
    assert env.render(mode="ansi").strip() == """
+---+---+---+---+---+
| 0 | 0 | 0 | 0 | 0 |
+---+---+---+---+---+
| 0 | 0 | 2 | 0 | 0 |
+---+---+---+---+---+
| 0 | 1 | 2 | 1 | 0 |
+---+---+---+---+---+
| 1 | 2 | 1 | 2 | 1 |
+---+---+---+---+---+
""".strip()


def test_can_run_agents():
    def custom1():
        return 1

    def custom2():
        return 2
    before_each(
        configuration={"rows": 4, "columns": 5, "inarow": 3},
    )
    board = [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 2, 0, 0, 0, 1, 2, 0, 0]
    assert env.run([custom1, custom2])[-1] == [
        {
            "action": 1,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "board": board, "mark": 1, "step": 5},
            "reward": 1,
        },
        {
            "action": 0,
            "status": "DONE",
            "info": {},
            "observation": {"remainingOverageTime": 60, "mark": 2},
            "reward": -1,
        },
    ]


def test_can_evaluate():
    rewards = evaluate("connectx", ["random", "random"], num_episodes=2)
    assert (rewards[0][0] + rewards[0][1] ==
            0) and rewards[1][0] + rewards[1][1] == 0



import json
from os import path
dir_path = path.dirname(__file__)
constants_path = path.abspath(path.join(dir_path, "game_constants.json"))
with open(constants_path) as f:
    GAME_CONSTANTS = json.load(f)


# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import kaggle_environments.helpers
from enum import auto, Enum
from kaggle_environments.helpers import histogram, with_print
from os import path
from random import choice, sample
from typing import *


class Observation(kaggle_environments.helpers.Observation):
    @property
    def geese(self) -> List[List[int]]:
        return self["geese"]

    @property
    def food(self) -> List[int]:
        return self["food"]

    @property
    def index(self) -> int:
        return self["index"]


class Configuration(kaggle_environments.helpers.Configuration):
    @property
    def columns(self) -> int:
        return self["columns"]

    @property
    def rows(self) -> int:
        return self["rows"]

    @property
    def hunger_rate(self) -> int:
        return self["hunger_rate"]

    @property
    def min_food(self) -> int:
        return self["min_food"]

    @property
    def max_length(self) -> int:
        return self["max_length"]


class Action(Enum):
    NORTH = auto()
    EAST = auto()
    SOUTH = auto()
    WEST = auto()

    def to_row_col(self):
        if self == Action.NORTH:
            return -1, 0
        if self == Action.SOUTH:
            return 1, 0
        if self == Action.EAST:
            return 0, 1
        if self == Action.WEST:
            return 0, -1
        return 0, 0

    def opposite(self):
        if self == Action.NORTH:
            return Action.SOUTH
        if self == Action.SOUTH:
            return Action.NORTH
        if self == Action.EAST:
            return Action.WEST
        if self == Action.WEST:
            return Action.EAST
        raise TypeError(str(self) + " is not a valid Action.")


def row_col(position: int, columns: int) -> Tuple[int, int]:
    return position // columns, position % columns


def translate(position: int, direction: Action, columns: int, rows: int) -> int:
    row, column = row_col(position, columns)
    row_offset, column_offset = direction.to_row_col()
    row = (row + row_offset) % rows
    column = (column + column_offset) % columns
    return row * columns + column


def adjacent_positions(position: int, columns: int, rows: int) -> List[int]:
    return [
        translate(position, action, columns, rows)
        for action in Action
    ]


def min_distance(position: int, food: List[int], columns: int):
    row, column = row_col(position, columns)
    return min(
        abs(row - food_row) + abs(column - food_column)
        for food_position in food
        for food_row, food_column in [row_col(food_position, columns)]
    )


def random_agent():
    return choice([action for action in Action]).name


class GreedyAgent:
    def __init__(self, configuration: Configuration):
        self.configuration = configuration
        self.last_action = None

    def __call__(self, observation: Observation):
        rows, columns = self.configuration.rows, self.configuration.columns

        food = observation.food
        geese = observation.geese
        opponents = [
            goose
            for index, goose in enumerate(geese)
            if index != observation.index and len(goose) > 0
        ]

        # Don't move adjacent to any heads
        head_adjacent_positions = {
            opponent_head_adjacent
            for opponent in opponents
            for opponent_head in [opponent[0]]
            for opponent_head_adjacent in adjacent_positions(opponent_head, columns, rows)
        }
        # Don't move into any bodies
        bodies = {position for goose in geese for position in goose}

        # Move to the closest food
        position = geese[observation.index][0]
        actions = {
            action: min_distance(new_position, food, columns)
            for action in Action
            for new_position in [translate(position, action, columns, rows)]
            if (
                new_position not in head_adjacent_positions and
                new_position not in bodies and
                (self.last_action is None or action != self.last_action.opposite())
            )
        }

        action = min(actions, key=actions.get) if any(actions) else choice([action for action in Action])
        self.last_action = action
        return action.name


cached_greedy_agents = {}


def greedy_agent(obs, config):
    index = obs["index"]
    if index not in cached_greedy_agents:
        cached_greedy_agents[index] = GreedyAgent(Configuration(config))
    return cached_greedy_agents[index](Observation(obs))


agents = {"random": random_agent, "greedy": greedy_agent}


def interpreter(state, env):
    configuration = Configuration(env.configuration)
    columns = configuration.columns
    rows = configuration.rows
    min_food = configuration.min_food
    state[0].observation = shared_observation = Observation(state[0].observation)

    # Reset the environment.
    if env.done:
        agent_count = len(state)
        heads = sample(range(columns * rows), agent_count)
        shared_observation["geese"] = [[head] for head in heads]
        food_candidates = set(range(columns * rows)).difference(heads)
        # Ensure we only place as many food as there are open squares
        min_food = min(min_food, len(food_candidates))
        shared_observation["food"] = sample(food_candidates, min_food)
        return state

    geese = shared_observation.geese
    food = shared_observation.food

    # If there is no last state, reuse current state so that current action is never the opposite of the last action.
    last_state = env.steps[-1] if len(env.steps) > 1 else state
    # Apply the actions from active agents.
    for index, agent in enumerate(state):
        if agent.status != "ACTIVE":
            if agent.status != "INACTIVE" and agent.status != "DONE":
                # ERROR, INVALID, or TIMEOUT, remove the goose.
                geese[index] = []
            continue

        action = Action[agent.action]

        # Check action direction
        last_agent = last_state[index]
        last_action = Action[last_agent["action"]] if "action" in last_agent else action
        if last_action == action.opposite():
            env.debug_print(f"Opposite action: {agent.observation.index, action, last_action}")
            agent.status = "DONE"
            geese[index] = []
            continue

        goose = geese[index]
        head = translate(goose[0], action, columns, rows)

        # Consume food or drop a tail piece.
        if head in food:
            food.remove(head)
        else:
            goose.pop()

        # Self collision.
        if head in goose:
            env.debug_print(f"Body Hit: {agent.observation.index, action, head, goose}")
            agent.status = "DONE"
            geese[index] = []
            continue

        while len(goose) >= configuration.max_length:
            # Free a spot for the new head if needed
            goose.pop()
        # Add New Head to the Goose.
        goose.insert(0, head)

        # If hunger strikes remove from the tail.
        if len(env.steps) % configuration.hunger_rate == 0:
            if len(goose) > 0:
                goose.pop()
            if len(goose) == 0:
                env.debug_print(f"Goose Starved: {action}")
                agent.status = "DONE"
                continue

    goose_positions = histogram(
        position
        for goose in geese
        for position in goose
    )

    # Check for collisions.
    for index, agent in enumerate(state):
        goose = geese[index]
        if len(goose) > 0:
            head = geese[index][0]
            if goose_positions[head] > 1:
                env.debug_print(f"Goose Collision: {agent.action}")
                agent.status = "DONE"
                geese[index] = []

    # Add food if min_food threshold reached.
    needed_food = min_food - len(food)
    if needed_food > 0:
        collisions = {
            position
            for goose in geese
            for position in goose
        }
        available_positions = set(range(rows * columns)).difference(collisions).difference(food)
        # Ensure we don't sample more food than available positions.
        needed_food = min(needed_food, len(available_positions))
        food.extend(sample(available_positions, needed_food))

    # Set rewards after deleting all geese to ensure that geese don't receive a reward on the turn they perish.
    for index, agent in enumerate(state):
        if agent.status == "ACTIVE":
            # Adding 1 to len(env.steps) ensures that if an agent gets reward 4507, it died on turn 45 with length 7.
            agent.reward = (len(env.steps) + 1) * (configuration.max_length + 1) + len(geese[index])

    # If only one ACTIVE agent left, set it to DONE.
    active_agents = [a for a in state if a.status == "ACTIVE"]
    if len(active_agents) == 1:
        agent = active_agents[0]
        agent.status = "DONE"

    return state


def renderer(state, env):
    config = env.configuration
    columns = config.columns
    rows = config.rows

    food_symbol = "F"
    column_divider = "|"
    row_divider = "+" + "+".join(["---"] * columns) + "+\n"

    board = [" "] * (rows * columns)
    for pos in state[0].observation.food:
        board[pos] = food_symbol

    for index, goose in enumerate(state[0].observation.geese):
        for position in goose:
            board[position] = index

    out = row_divider
    for row in range(rows):
        for col in range(columns):
            out += column_divider + f" {board[(row * columns) + col]} "
        out += column_divider + "\n" + row_divider

    return out


dirpath = path.dirname(__file__)
jsonpath = path.abspath(path.join(dirpath, "hungry_geese.json"))
with open(jsonpath) as f:
    specification = json.load(f)


def html_renderer():
    jspath = path.abspath(path.join(dirpath, "hungry_geese.js"))
    with open(jspath) as f:
        return f.read()



# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import traceback
import copy
import json
import uuid
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO
from multiprocessing import Pool
from time import perf_counter
from .agent import Agent
from .errors import DeadlineExceeded, FailedPrecondition, Internal, InvalidArgument
from .utils import get, has, get_player, process_schema, schemas, structify

# Registered Environments.
environments = {}

# Registered Interactive Sessions.
interactives = {}


def register(name, environment):
    """
    Register an environment by name.  An environment contains the following:
     * specification - JSON Schema representing the environment.
     * interpreter - Function(state, environment) -> new_state
     * renderer - Function(state, environment) -> string
     * html_renderer - Function(environment) -> JavaScript HTML renderer function.
     * agents(optional) - List of default agents [Function(observation, config) -> action]
    """
    environments[name] = environment


def evaluate(environment, agents=[], configuration={}, steps=[], num_episodes=1, debug=False, state=None):
    """
    Evaluate and return the rewards of one or more episodes (environment and agents combo).

    Args:
        environment (str|Environment):
        agents (list):
        configuration (dict, optional):
        steps (list, optional):
        num_episodes (int=1, optional): How many episodes to execute (run until done).
        debug (bool=False, optional): Render print() statments to stdout
        state (optional)

    Returns:
        list of list of int: List of final rewards for all agents for all episodes.
    """
    e = make(environment, configuration, steps, debug=debug, state=state)
    rewards = [[]] * num_episodes
    for i in range(num_episodes):
        last_state = e.run(agents)[-1]
        rewards[i] = [state.reward for state in last_state]
    return rewards


def make(environment, configuration={}, info={}, steps=[], logs=[], debug=False, state=None):
    """
    Creates an instance of an Environment.

    Args:
        environment (str|Environment):
        configuration (dict, optional):
        info (dict, optional):
        steps (list, optional):
        debug (bool=False, optional): Render print() statments to stdout
        state (optional):

    Returns:
        Environment: Instance of a specific environment.
    """
    if has(environment, str) and has(environments, dict, path=[environment]):
        return Environment(**environments[environment], configuration=configuration, info=info, steps=steps, logs=logs, debug=debug, state=state)
    elif callable(environment):
        return Environment(interpreter=environment, configuration=configuration, info=info, steps=steps, logs=logs, debug=debug, state=state)
    elif has(environment, path=["interpreter"], is_callable=True):
        return Environment(**environment, configuration=configuration, info=info, steps=steps, logs=logs, debug=debug, state=state)
    raise InvalidArgument("Unknown Environment Specification")


def act_agent(args):
    agent, state, configuration, none_action = args
    if state["status"] != "ACTIVE":
        return None, {}
    elif agent is None:
        return none_action, {}
    else:
        return agent.act(state["observation"])


class Environment:
    def __init__(
        self,
        specification={},
        configuration={},
        info={},
        steps=[],
        logs=[],
        agents={},
        interpreter=None,
        renderer=None,
        html_renderer=None,
        debug=False,
        state=None,
    ):
        self.logs = logs
        self.id = str(uuid.uuid1())
        self.debug = debug
        self.info = info
        self.pool = None

        err, specification = self.__process_specification(specification)
        if err:
            raise InvalidArgument("Specification Invalid: " + err)
        self.specification = structify(specification)

        err, configuration = process_schema(
            {"type": "object", "properties": self.specification.configuration},
            {} if configuration is None else configuration,
        )
        if err:
            raise InvalidArgument("Configuration Invalid: " + err)
        self.configuration = structify(configuration)

        if not callable(interpreter):
            raise InvalidArgument("Interpreter is not Callable.")
        self.interpreter = interpreter

        if not callable(renderer):
            raise InvalidArgument("Renderer is not Callable.")
        self.renderer = renderer

        if not callable(html_renderer):
            raise InvalidArgument("Html_renderer is not Callable.")
        self.html_renderer = html_renderer

        if not all([callable(a) for a in agents.values()]):
            raise InvalidArgument("Default agents must be Callable.")
        self.agents = structify(agents)

        if steps is not None and len(steps) > 0:
            self.__set_state(steps[-1])
            self.steps = steps[0:-1] + self.steps
        elif state is not None:
            step = [{}] * self.specification.agents[0]
            step[0] = state
            self.__set_state(step)
        else:
            self.reset()

    def step(self, actions, logs=[]):
        """
        Execute the environment interpreter using the current state and a list of actions.

        Args:
            actions (list): Actions to pair up with the current agent states.
            logs (list): Logs to pair up with each agent for the current step.

        Returns:
            list of dict: The agents states after the step.
        """
        if self.done:
            raise FailedPrecondition("Environment done, reset required.")
        if not actions or len(actions) != len(self.state):
            raise InvalidArgument(f"{len(self.state)} actions required.")

        action_state = [0] * len(self.state)
        for index, action in enumerate(actions):
            action_state[index] = {**self.state[index], "action": None}

            if isinstance(action, DeadlineExceeded):
                self.debug_print(f"Timeout: {str(action)}")
                action_state[index]["status"] = "TIMEOUT"
            elif isinstance(action, BaseException):
                self.debug_print(f"Error: {traceback.format_exception(None, action, action.__traceback__)}")
                action_state[index]["status"] = "ERROR"
            else:
                err, data = process_schema(
                    self.__state_schema.properties.action, action)
                if err:
                    self.debug_print(f"Invalid Action: {str(err)}")
                    action_state[index]["status"] = "INVALID"
                else:
                    action_state[index]["action"] = data

        self.state = self.__run_interpreter(action_state, logs)

        # Max Steps reached. Mark ACTIVE/INACTIVE agents as DONE.
        if self.state[0].observation.step >= self.configuration.episodeSteps - 1:
            for s in self.state:
                if s.status == "ACTIVE" or s.status == "INACTIVE":
                    s.status = "DONE"

        self.steps.append(self.state)
        if logs is not None:
            self.logs.append(logs)

        return self.state

    def run(self, agents):
        """
        Steps until the environment is "done" or the runTimeout was reached.

        Args:
            agents (list of any): List of agents to obtain actions from.

        Returns:
            tuple of:
                list of list of dict: The agent states of all steps executed.
                list of list of dict: The agent logs of all steps executed.
        """
        if self.state is None or len(self.steps) == 1 or self.done:
            self.reset(len(agents))
        if len(self.state) != len(agents):
            raise InvalidArgument(
                f"{len(self.state)} agents were expected, but {len(agents)} was given.")

        runner = self.__agent_runner(agents)
        start = perf_counter()
        while not self.done and perf_counter() - start < self.configuration.runTimeout:
            actions, logs = runner.act()
            self.step(actions, logs)
        return self.steps

    def reset(self, num_agents=None):
        """
        Resets the environment state to the initial step.

        Args:
            num_agents (int): Resets the state assuming a fixed number of agents.

        Returns:
            list of dict: The agents states after the reset.
        """

        if num_agents is None:
            num_agents = self.specification.agents[0]

        # Get configuration default state.
        self.__set_state([{} for _ in range(num_agents)])
        # Reset all agents to status=INACTIVE (copy out values to reset afterwards).
        statuses = [a.status for a in self.state]
        for agent in self.state:
            agent.status = "INACTIVE"
        # Give the interpreter an opportunity to make any initializations.
        logs = []
        self.__set_state(self.__run_interpreter(self.state, logs))
        self.logs.append(logs)
        # Replace the starting "status" if still "done".
        if self.done and len(self.state) == len(statuses):
            for i in range(len(self.state)):
                self.state[i].status = statuses[i]
        return self.state

    def render(self, **kwargs):
        """
        Renders a visual representation of the current state of the environment.

        Args:
            mode (str): html, ipython, ansi, human (default)
            **kwargs (dict): Other args are directly passed into the html player.

        Returns:
            str: html if mode=html or ansi if mode=ansi.
            None: prints ansi if mode=human or prints html if mode=ipython
        """
        mode = get(kwargs, str, "human", path=["mode"])
        if mode == "ansi" or mode == "human":
            args = [self.state, self]
            out = self.renderer(*args[:self.renderer.__code__.co_argcount])
            if mode == "ansi":
                return out
        elif mode == "html" or mode == "ipython":
            is_playing = get(kwargs, bool, self.done, path=["playing"])
            window_kaggle = {
                "debug": get(kwargs, bool, self.debug, path=["debug"]),
                "playing": is_playing,
                "step": 0 if is_playing else len(self.steps) - 1,
                "controls": get(kwargs, bool, self.done, path=["controls"]),
                "environment": self.toJSON(),
                "logs": self.logs,
                **kwargs,
            }
            args = [self]
            player_html = get_player(window_kaggle,
                                     self.html_renderer(*args[:self.html_renderer.__code__.co_argcount]))
            if mode == "html":
                return player_html

            from IPython.display import display, HTML
            player_html = player_html.replace('"', '&quot;')
            width = get(kwargs, int, 300, path=["width"])
            height = get(kwargs, int, 300, path=["height"])
            html = f'<iframe srcdoc="{player_html}" width="{width}" height="{height}" frameborder="0"></iframe> '
            display(HTML(html))
        elif mode == "json":
            return json.dumps(self.toJSON(), sort_keys=True, indent=2 if self.debug else None)
        else:
            raise InvalidArgument("Available render modes: human, ansi, html, ipython")

    def play(self, agents=[], **kwargs):
        """
        Renders a visual representation of the environment and allows interactive action selection.

        Args:
            **kwargs (dict): Args directly passed into render().  Mode is fixed to ipython.

        Returns:
            None: prints directly to an IPython notebook
        """
        env = self.clone()
        trainer = env.train(agents)
        interactives[env.id] = (env, trainer)
        env.render(mode="ipython", interactive=True, **kwargs)

    def train(self, agents=[]):
        """
        Setup a lightweight training environment for a single agent.
        Note: This is designed to be a lightweight starting point which can
              be integrated with other frameworks (i.e. gym, stable-baselines).
              The reward returned by the "step" function here is a diff between the
              current and the previous step.

        Example:
            env = make("tictactoe")
            # Training agent in first position (player 1) against the default random agent.
            trainer = env.train([None, "random"])

            obs = trainer.reset()
            done = False
            while not done:
                action = 0 # Action for the agent being trained.
                obs, reward, done, info = trainer.step(action)
            env.render()

        Args:
            agents (list): List of agents to obtain actions from while training.
                           The agent to train (in position), should be set to "None".

        Returns:
            `dict`.reset: Reset def that reset the environment, then advances until the agents turn.
            `dict`.step: Steps using the agent action, then advance until agents turn again.
        """
        runner = None
        position = None
        for index, agent in enumerate(agents):
            if agent is None:
                if position is not None:
                    raise InvalidArgument(
                        "Only one agent can be marked 'None'")
                position = index

        if position is None:
            raise InvalidArgument("One agent must be marked 'None' to train.")

        def advance():
            while not self.done and self.state[position].status == "INACTIVE":
                actions, logs = runner.act()
                self.step(actions, logs)

        def reset():
            nonlocal runner
            self.reset(len(agents))
            runner = self.__agent_runner(agents)
            advance()
            return self.__get_shared_state(position).observation

        def step(action):
            actions, logs = runner.act(action)
            self.step(actions, logs)
            advance()
            agent = self.__get_shared_state(position)
            reward = agent.reward
            if len(self.steps) > 1 and reward is not None:
                reward -= self.steps[-2][position].reward
            return [
                agent.observation, reward, agent.status != "ACTIVE", agent.info
            ]

        reset()

        return structify({"step": step, "reset": reset})

    @property
    def name(self):
        """str: The name from the specification."""
        return get(self.specification, str, "", ["name"])

    @property
    def version(self):
        """str: The version from the specification."""
        return get(self.specification, str, "", ["version"])

    @property
    def done(self):
        """bool: If any agents have an ACTIVE status."""
        return all(s.status != "ACTIVE" for s in self.state)

    def toJSON(self):
        """
        Returns:
            dict: Specifcation and current state of the Environment instance.
        """
        spec = self.specification
        return copy.deepcopy(
            {
                "id": self.id,
                "name": spec.name,
                "title": spec.title,
                "description": spec.description,
                "version": spec.version,
                "configuration": self.configuration,
                "specification": {
                    "action": spec.action,
                    "agents": spec.agents,
                    "configuration": spec.configuration,
                    "info": spec.info,
                    "observation": spec.observation,
                    "reward": spec.reward
                },
                "steps": self.steps,
                "rewards": [state.reward for state in self.steps[-1]],
                "statuses": [state.status for state in self.steps[-1]],
                "schema_version": 1,
                "info": self.info,
            }
        )

    def clone(self):
        """
        Returns:
            Environment: A copy of the current environment.
        """
        return Environment(
            specification=self.specification,
            configuration=self.configuration,
            steps=self.steps,
            agents=self.agents,
            interpreter=self.interpreter,
            renderer=self.renderer,
            html_renderer=self.html_renderer,
            debug=self.debug,
        )

    @property
    def __state_schema(self):
        if not hasattr(self, "__state_schema_value"):
            spec = self.specification
            self.__state_schema_value = {
                **schemas["state"],
                "properties": {
                    "action": {
                        **schemas.state.properties.action,
                        **get(spec, dict, path=["action"], fallback={})
                    },
                    "reward": {
                        **schemas.state.properties.reward,
                        **get(spec, dict, path=["reward"], fallback={})
                    },
                    "info": {
                        **schemas.state.properties.info,
                        "properties": get(spec, dict, path=["info"], fallback={})
                    },
                    "observation": {
                        **schemas.state.properties.observation,
                        "properties": get(spec, dict, path=["observation"], fallback={})
                    },
                    "status": {
                        **schemas.state.properties.status,
                        **get(spec, dict, path=["status"], fallback={})
                    },
                },
            }
        return structify(self.__state_schema_value)

    def __set_state(self, state=[]):
        if len(state) not in self.specification.agents:
            raise InvalidArgument(
                f"{len(state)} is not a valid number of agent(s).")

        self.state = structify([self.__get_state(index, s)
                                for index, s in enumerate(state)])
        self.steps = [self.state]
        return self.state

    def __get_state(self, position, state):
        key = f"__state_schema_{position}"
        if not hasattr(self, key):

            # Update a property default value based on position in defaults.
            # Remove shared properties from non-first agents.
            def update_props(props):
                for k, prop in list(props.items()):
                    if get(prop, bool, path=["shared"], fallback=False) and position > 0:
                        del props[k]
                        continue
                    if has(prop, list, path=["defaults"]) and len(prop["defaults"]) > position:
                        prop["default"] = prop["defaults"][position]
                        del prop["defaults"]
                    if has(prop, dict, path=["properties"]):
                        update_props(prop["properties"])
                return props

            props = structify(update_props(
                copy.deepcopy(self.__state_schema.properties)))

            setattr(self, key, {**self.__state_schema, "properties": props})

        err, data = process_schema(getattr(self, key), state)
        if err:
            raise InvalidArgument(
                f"Default state generation failed for #{position}: " + err
            )
        return data

    def __run_interpreter(self, state, logs):
        out = None
        err = None
        # Append any environmental logs to any agent logs we collected.
        try:
            with StringIO() as out_buffer, StringIO() as err_buffer, redirect_stdout(out_buffer), redirect_stderr(err_buffer):
                try:
                    args = [structify(state), self]
                    new_state = structify(self.interpreter(
                        *args[:self.interpreter.__code__.co_argcount]))
                    new_state[0].observation.step = (
                        0 if self.done
                        else len(self.steps)
                    )

                    for index, agent in enumerate(new_state):
                        if index < len(logs) and "duration" in logs[index]:
                            duration = logs[index]["duration"]
                            overage_time_consumed = max(0, duration - self.configuration.actTimeout)
                            agent.observation.remainingOverageTime -= overage_time_consumed
                        if agent.status not in self.__state_schema.properties.status.enum:
                            self.debug_print(f"Invalid Action: {agent.status}")
                            agent.status = "INVALID"
                        if agent.status in ["ERROR", "INVALID", "TIMEOUT"]:
                            agent.reward = None
                    return new_state
                except Exception as e:
                    # Print the exception stack trace to our log
                    traceback.print_exc(file=err_buffer)
                    # Reraise e to ensure that the program exits
                    raise e
                finally:
                    # Allow up to 1k log characters per step which is ~1MB per 600 step episode
                    max_log_length = 1024
                    out = out_buffer.getvalue()
                    err = err_buffer.getvalue()
                    if out or err:
                        logs.append({
                            "stdout": out[0:max_log_length],
                            "stderr": err[0:max_log_length]
                        })
        finally:
            if out:
                while out.endswith('\n'):
                    out = out[:-1]
                self.debug_print(out)
            if err:
                while err.endswith('\n'):
                    err = err[:-1]
                self.debug_print(err)

    def __process_specification(self, spec):
        if has(spec, path=["reward"]):
            reward = spec["reward"]
            reward_type = get(reward, str, "number", ["type"])
            if reward_type not in ["integer", "number"]:
                return ("type must be an integer or number", None)
            reward["type"] = [reward_type, "null"]

        # Allow environments to extend various parts of the specification.
        def extend_specification(source, field_name):
            field = copy.deepcopy(source[field_name]["properties"])
            for k, v in get(spec, dict, {}, [field_name]).items():
                # Set a new default value.
                if not isinstance(v, dict):
                    if not has(field, path=[k]):
                        raise InvalidArgument(
                            f"Field {field} was unable to set default of missing property: {k}")
                    field[k]["default"] = v
                # Add a new field.
                elif not has(field, path=[k]):
                    field[k] = v
                # Override an existing field if types match.
                elif field[k]["type"] == get(v, path=["type"]):
                    field[k] = v
                # Types don't match - unable to extend.
                else:
                    raise InvalidArgument(
                        f"Field {field} was unable to extend: {k}")

            spec[field_name] = field

        extend_specification(schemas, "configuration")
        extend_specification(schemas["state"]["properties"], "observation")

        return process_schema(schemas.specification, spec)

    def __agent_runner(self, agents):
        # Generate the agents.
        agents = [
            Agent(agent, self)
            if agent is not None
            else None
            for agent in agents
        ]

        def act(none_action=None):
            if len(agents) != len(self.state):
                raise InvalidArgument(
                    "Number of agents must match the state length")

            act_args = [
                (
                    agent,
                    self.__get_shared_state(i),
                    self.configuration,
                    none_action,
                )
                for i, agent in enumerate(agents)
            ]

            if all((agent is None or agent.is_parallelizable) for agent in agents):
                if self.pool is None:
                    self.pool = Pool(processes=len(agents))
                results = self.pool.map(act_agent, act_args)
            else:
                results = list(map(act_agent, act_args))

            # results is a list of tuples where the first element is an agent action and the second is the agent log
            # This destructures into two lists, a list of actions and a list of logs.
            actions, logs = zip(*results)
            return list(actions), list(logs)

        return structify({"act": act})

    def __get_shared_state(self, position):
        # Note: state and schema are required to be in sync (apart from shared ones).
        def update_props(shared_state, state, schema_props):
            for k, prop in schema_props.items():
                # Hidden fields are tracked in the episode replay but are not provided to the agent at runtime
                if get(prop, bool, path=["hidden"], fallback=False):
                    if k in state:
                        del state[k]
                elif get(prop, bool, path=["shared"], fallback=False):
                    state[k] = shared_state[k]
                elif has(prop, dict, path=["properties"]):
                    update_props(shared_state[k], state[k], prop["properties"])
            return state

        return update_props(
            self.state[0],
            copy.deepcopy(self.state[position]),
            self.__state_schema.properties
        )

    def debug_print(self, message):
        if self.debug:
            print(message)



# Copyright 2020 Kaggle Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from enum import Enum, auto
from functools import wraps
from kaggle_environments.helpers import Point, group_by
from typing import *
import sys
import kaggle_environments.helpers


# region Data Model Classes
class Observation(kaggle_environments.helpers.Observation):
    """
    Observation primarily used as a helper to construct the Board from the raw observation.
    This provides bindings for the observation type described at https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/halite/halite.json
    """
    @property
    def halite(self) -> List[float]:
        """Serialized list of available halite per cell on the board."""
        return self["halite"]

    @property
    def players(self) -> List[List[int]]:
        """List of players and their assets."""
        return self["players"]

    @property
    def player(self) -> int:
        """The current agent's player index."""
        return self["player"]


class Configuration(kaggle_environments.helpers.Configuration):
    """
    Configuration provides access to tunable parameters in the environment.
    This provides bindings for the configuration type described at https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/halite/halite.json
    """
    @property
    def agent_timeout(self) -> float:
        """Maximum runtime (seconds) to initialize an agent."""
        return self["agentTimeout"]

    @property
    def starting_halite(self) -> int:
        """The starting amount of halite available on the board."""
        return self["startingHalite"]

    @property
    def size(self) -> int:
        """The number of cells vertically and horizontally on the board."""
        return self["size"]

    @property
    def spawn_cost(self) -> int:
        """The amount of halite to spawn a new ship."""
        return self["spawnCost"]

    @property
    def convert_cost(self) -> int:
        """The amount of halite to convert a ship into a shipyard."""
        return self["convertCost"]

    @property
    def move_cost(self) -> float:
        """The percent deducted from ship's current halite per move."""
        return self["moveCost"]

    @property
    def collect_rate(self) -> float:
        """The rate of halite collected by a ship from a cell by not moving."""
        return self["collectRate"]

    @property
    def regen_rate(self) -> float:
        """The rate halite regenerates on the board."""
        return self["regenRate"]

    @property
    def max_cell_halite(self) -> int:
        """The maximum halite that can be in any cell."""
        return self["maxCellHalite"]

    @property
    def random_seed(self) -> int:
        """The seed to the random number generator (0 means no seed)."""
        return self["randomSeed"]


class ShipAction(Enum):
    NORTH = auto()
    EAST = auto()
    SOUTH = auto()
    WEST = auto()
    CONVERT = auto()
    # Use None to collect halite on a cell

    def to_point(self) -> Optional[Point]:
        """
        This returns the position offset associated with a particular action or None if the action does not change the ship's position.
        NORTH -> (0, 1)
        EAST -> (1, 0)
        SOUTH -> (0, -1)
        WEST -> (-1, 0)
        """
        return (
            Point(0, 1) if self == ShipAction.NORTH else
            Point(1, 0) if self == ShipAction.EAST else
            Point(0, -1) if self == ShipAction.SOUTH else
            Point(-1, 0) if self == ShipAction.WEST else
            None
        )

    def __str__(self) -> str:
        return self.name

    @staticmethod
    def moves() -> List['ShipAction']:
        return [
            ShipAction.NORTH,
            ShipAction.EAST,
            ShipAction.SOUTH,
            ShipAction.WEST,
        ]


class ShipyardAction(Enum):
    SPAWN = auto()

    def __str__(self) -> str:
        return self.name


ShipId = NewType('ShipId', str)
ShipyardId = NewType('ShipyardId', str)
PlayerId = NewType('PlayerId', int)


class Cell:
    def __init__(self, position: Point, halite: float, shipyard_id: Optional[ShipyardId], ship_id: Optional[ShipId], board: 'Board') -> None:
        self._position = position
        self._halite = halite
        self._shipyard_id = shipyard_id
        self._ship_id = ship_id
        self._board = board

    @property
    def position(self) -> Point:
        return self._position

    @property
    def halite(self) -> float:
        return self._halite

    @property
    def shipyard_id(self) -> Optional[ShipyardId]:
        return self._shipyard_id

    @property
    def ship_id(self) -> Optional[ShipId]:
        return self._ship_id

    @property
    def ship(self) -> Optional['Ship']:
        """Returns the ship on this cell if it exists and None otherwise."""
        return self._board.ships.get(self.ship_id)

    @property
    def shipyard(self) -> Optional['Shipyard']:
        """Returns the shipyard on this cell if it exists and None otherwise."""
        return self._board.shipyards.get(self.shipyard_id)

    def neighbor(self, offset: Point) -> 'Cell':
        """Returns the cell at self.position + offset."""
        (x, y) = self.position + offset
        return self._board[x, y]

    @property
    def north(self) -> 'Cell':
        """Returns the cell north of this cell."""
        return self.neighbor(ShipAction.NORTH.to_point())

    @property
    def south(self) -> 'Cell':
        """Returns the cell south of this cell."""
        return self.neighbor(ShipAction.SOUTH.to_point())

    @property
    def east(self) -> 'Cell':
        """Returns the cell east of this cell."""
        return self.neighbor(ShipAction.EAST.to_point())

    @property
    def west(self) -> 'Cell':
        """Returns the cell west of this cell."""
        return self.neighbor(ShipAction.WEST.to_point())


class Ship:
    def __init__(self, ship_id: ShipId, position: Point, halite: int, player_id: PlayerId, board: 'Board', next_action: Optional[ShipAction] = None) -> None:
        self._id = ship_id
        self._position = position
        self._halite = halite
        self._player_id = player_id
        self._board = board
        self._next_action = next_action

    @property
    def id(self) -> ShipId:
        return self._id

    @property
    def position(self) -> Point:
        return self._position

    @property
    def halite(self) -> int:
        return self._halite

    @property
    def player_id(self) -> PlayerId:
        return self._player_id

    @property
    def cell(self) -> Cell:
        """Returns the cell this ship is on."""
        return self._board[self.position]

    @property
    def player(self) -> 'Player':
        """Returns the player that owns this ship."""
        return self._board.players[self.player_id]

    @property
    def next_action(self) -> Optional[ShipAction]:
        """Returns the action that will be executed by this ship when Board.next() is called (when the current turn ends)."""
        return self._next_action

    @next_action.setter
    def next_action(self, value: Optional[ShipAction]) -> None:
        """Sets the action that will be executed by this ship when Board.next() is called (when the current turn ends)."""
        self._next_action = value

    @property
    def _observation(self) -> List[int]:
        """Converts a ship back to the normalized observation subset that constructed it."""
        return [self.position.to_index(self._board.configuration.size), self.halite]


class Shipyard:
    def __init__(self, shipyard_id: ShipyardId, position: Point, player_id: PlayerId, board: 'Board', next_action: Optional[ShipyardAction] = None) -> None:
        self._id = shipyard_id
        self._position = position
        self._player_id = player_id
        self._board = board
        self._next_action = next_action

    @property
    def id(self) -> ShipyardId:
        return self._id

    @property
    def position(self) -> Point:
        return self._position

    @property
    def player_id(self) -> PlayerId:
        return self._player_id

    @property
    def cell(self) -> Cell:
        """Returns the cell this shipyard is on."""
        return self._board[self.position]

    @property
    def player(self) -> 'Player':
        return self._board.players[self.player_id]

    @property
    def next_action(self) -> ShipyardAction:
        """Returns the action that will be executed by this shipyard when Board.next() is called (when the current turn ends)."""
        return self._next_action

    @next_action.setter
    def next_action(self, value: Optional[ShipyardAction]) -> None:
        """Sets the action that will be executed by this shipyard when Board.next() is called (when the current turn ends)."""
        self._next_action = value

    @property
    def _observation(self) -> int:
        """Converts a shipyard back to the normalized observation subset that constructed it."""
        return self.position.to_index(self._board.configuration.size)


class Player:
    def __init__(self, player_id: PlayerId, halite: int, shipyard_ids: List[ShipyardId], ship_ids: List[ShipId], board: 'Board') -> None:
        self._id = player_id
        self._halite = halite
        self._shipyard_ids = shipyard_ids
        self._ship_ids = ship_ids
        self._board = board

    @property
    def id(self) -> PlayerId:
        return self._id

    @property
    def halite(self) -> int:
        return self._halite

    @property
    def shipyard_ids(self) -> List[ShipyardId]:
        return self._shipyard_ids

    @property
    def ship_ids(self) -> List[ShipId]:
        return self._ship_ids

    @property
    def shipyards(self) -> List[Shipyard]:
        """Returns all shipyards owned by this player."""
        return [
            self._board.shipyards[shipyard_id]
            for shipyard_id in self.shipyard_ids
        ]

    @property
    def ships(self) -> List[Ship]:
        """Returns all ships owned by this player."""
        return [
            self._board.ships[ship_id]
            for ship_id in self.ship_ids
        ]

    @property
    def is_current_player(self) -> bool:
        """Returns whether this player is the current player (generally if this returns True, this player is you)."""
        return self.id == self._board.current_player_id

    @property
    def next_actions(self) -> Dict[str, str]:
        """Returns all queued ship and shipyard actions for this player formatted for the halite interpreter to receive as an agent response."""
        ship_actions = {
            ship.id: ship.next_action.name
            for ship in self.ships
            if ship.next_action is not None
        }
        shipyard_actions = {
            shipyard.id: shipyard.next_action.name
            for shipyard in self.shipyards
            if shipyard.next_action is not None
        }
        return {**ship_actions, **shipyard_actions}

    @property
    def _observation(self):
        """Converts a player back to the normalized observation subset that constructed it."""
        shipyards = {shipyard.id: shipyard._observation for shipyard in self.shipyards}
        ships = {ship.id: ship._observation for ship in self.ships}
        return [self.halite, shipyards, ships]
# endregion


class Board:
    def __init__(
        self,
        raw_observation: Dict[str, Any],
        raw_configuration: Union[Configuration, Dict[str, Any]],
        next_actions: Optional[List[Dict[str, str]]] = None
    ) -> None:
        """
        Creates a board from the provided observation, configuration, and next_actions as specified by
        https://github.com/Kaggle/kaggle-environments/blob/master/kaggle_environments/envs/halite/halite.json
        Board tracks players (by id), ships (by id), shipyards (by id), and cells (by position).
        Each entity contains both key values (e.g. ship.player_id) as well as entity references (e.g. ship.player).
        References are deep and chainable e.g.
            [ship.halite for player in board.players for ship in player.ships]
            ship.player.shipyards[0].cell.north.east.ship
        Consumers should not set or modify any attributes except Ship.next_action and Shipyard.next_action
        """
        observation = Observation(raw_observation)
        # next_actions is effectively a Dict[Union[[ShipId, ShipAction], [ShipyardId, ShipyardAction]]]
        # but that type's not very expressible so we simplify it to Dict[str, str]
        # Later we'll iterate through it once for each ship and shipyard to pull all the actions out
        next_actions = next_actions or ([{}] * len(observation.players))

        self._step = observation.step
        self._remaining_overage_time = observation.remaining_overage_time
        self._configuration = Configuration(raw_configuration)
        self._current_player_id = observation.player
        self._players: Dict[PlayerId, Player] = {}
        self._ships: Dict[ShipId, Ship] = {}
        self._shipyards: Dict[ShipyardId, Shipyard] = {}
        self._cells: Dict[Point, Cell] = {}

        size = self.configuration.size
        # Create a cell for every point in a size x size grid
        for x in range(size):
            for y in range(size):
                position = Point(x, y)
                halite = observation.halite[position.to_index(size)]
                # We'll populate the cell's ships and shipyards in _add_ship and _add_shipyard
                self.cells[position] = Cell(position, halite, None, None, self)

        for (player_id, player_observation) in enumerate(observation.players):
            # We know the len(player_observation) == 3 based on the schema -- this is a hack to have a tuple in json
            [player_halite, player_shipyards, player_ships] = player_observation
            # We'll populate the player's ships and shipyards in _add_ship and _add_shipyard
            self.players[player_id] = Player(player_id, player_halite, [], [], self)
            player_actions = next_actions[player_id] or {}

            for (ship_id, [ship_index, ship_halite]) in player_ships.items():
                # In the raw observation, halite is stored as a 1d list but we convert it to a 2d dict for convenience
                # Accordingly we also need to convert our list indices to dict keys / 2d positions
                ship_position = Point.from_index(ship_index, size)
                raw_action = player_actions.get(ship_id)
                action = (
                    ShipAction[raw_action]
                    if raw_action in ShipAction.__members__
                    else None
                )
                self._add_ship(Ship(ship_id, ship_position, ship_halite, player_id, self, action))

            for (shipyard_id, shipyard_index) in player_shipyards.items():
                shipyard_position = Point.from_index(shipyard_index, size)
                raw_action = player_actions.get(shipyard_id)
                action = (
                    ShipyardAction[raw_action]
                    if raw_action in ShipyardAction.__members__
                    else None
                )
                self._add_shipyard(Shipyard(shipyard_id, shipyard_position, player_id, self, action))

    @property
    def configuration(self) -> Configuration:
        return self._configuration

    @property
    def players(self) -> Dict[PlayerId, Player]:
        return self._players

    @property
    def ships(self) -> Dict[ShipId, Ship]:
        """Returns all ships on the current board."""
        return self._ships

    @property
    def shipyards(self) -> Dict[ShipyardId, Shipyard]:
        """Returns all shipyards on the current board."""
        return self._shipyards

    @property
    def cells(self) -> Dict[Point, Cell]:
        """Returns all cells on the current board."""
        return self._cells

    @property
    def step(self) -> int:
        return self._step

    @property
    def current_player_id(self) -> PlayerId:
        return self._current_player_id

    @property
    def current_player(self) -> Player:
        """Returns the current player (generally this is you)."""
        return self._players[self.current_player_id]

    @property
    def opponents(self) -> List[Player]:
        """
        Returns all players that aren't the current player.
        You can get all opponent ships with [ship for ship in player.ships for player in board.opponents]
        """
        return [player for player in self.players.values() if not player.is_current_player]

    @property
    def observation(self) -> Dict[str, Any]:
        """Converts a Board back to the normalized observation that constructed it."""
        size = self.configuration.size
        halite = [self[Point.from_index(index, size)].halite for index in range(size * size)]
        players = [player._observation for player in self.players.values()]

        return {
            "halite": halite,
            "players": players,
            "player": self.current_player_id,
            "step": self.step,
            "remainingOverageTime": self._remaining_overage_time,
        }

    def __deepcopy__(self, _) -> 'Board':
        actions = [player.next_actions for player in self.players.values()]
        return Board(self.observation, self.configuration, actions)

    def __getitem__(self, point: Union[Tuple[int, int], Point]) -> Cell:
        """
        This method will wrap the supplied position to fit within the board size and return the cell at that location.
        e.g. on a 3x3 board, board[2, 1] is the same as board[5, 1]
        """
        if not isinstance(point, Point):
            (x, y) = point
            point = Point(x, y)
        return self._cells[point % self.configuration.size]

    def __str__(self) -> str:
        """
        The board is printed in a grid with the following rules:
        Capital letters are shipyards
        Lower case letters are ships
        Digits are cell halite and scale from 0-9 directly proportional to a value between 0 and self.configuration.max_cell_halite
        Player 1 is letter a/A
        Player 2 is letter b/B
        etc.
        """
        size = self.configuration.size
        result = ''
        for y in range(size):
            for x in range(size):
                cell = self[(x, size - y - 1)]
                result += '|'
                result += (
                    chr(ord('a') + cell.ship.player_id)
                    if cell.ship is not None
                    else ' '
                )
                # This normalizes a value from 0 to max_cell halite to a value from 0 to 9
                normalized_halite = int(9.0 * cell.halite / float(self.configuration.max_cell_halite))
                result += str(normalized_halite)
                result += (
                    chr(ord('A') + cell.shipyard.player_id)
                    if cell.shipyard is not None
                    else ' '
                )
            result += '|\n'
        return result

    def _add_ship(self: 'Board', ship: Ship) -> None:
        ship.player.ship_ids.append(ship.id)
        ship.cell._ship_id = ship.id
        self._ships[ship.id] = ship

    def _add_shipyard(self: 'Board', shipyard: Shipyard) -> None:
        shipyard.player.shipyard_ids.append(shipyard.id)
        shipyard.cell._shipyard_id = shipyard.id
        shipyard.cell._halite = 0
        self._shipyards[shipyard.id] = shipyard

    def _delete_ship(self: 'Board', ship: Ship) -> None:
        ship.player.ship_ids.remove(ship.id)
        if ship.cell.ship_id == ship.id:
            ship.cell._ship_id = None
        del self._ships[ship.id]

    def _delete_shipyard(self: 'Board', shipyard: Shipyard) -> None:
        shipyard.player.shipyard_ids.remove(shipyard.id)
        if shipyard.cell.shipyard_id == shipyard.id:
            shipyard.cell._shipyard_id = None
        del self._shipyards[shipyard.id]

    def next(self) -> 'Board':
        """
        Returns a new board with the current board's next actions applied.
        The current board is unmodified.
        This can form a halite interpreter, e.g.
            next_observation = Board(current_observation, configuration, actions).next().observation
        """
        # Create a copy of the board to modify so we don't affect the current board
        board = deepcopy(self)
        configuration = board.configuration
        convert_cost = configuration.convert_cost
        spawn_cost = configuration.spawn_cost
        uid_counter = 0

        # This is a consistent way to generate unique strings to form ship and shipyard ids
        def create_uid():
            nonlocal uid_counter
            uid_counter += 1
            return f"{self.step + 1}-{uid_counter}"

        # Process actions and store the results in the ships and shipyards lists for collision checking
        for player in board.players.values():
            leftover_convert_halite = 0

            for shipyard in player.shipyards:
                if shipyard.next_action == ShipyardAction.SPAWN and player.halite >= spawn_cost:
                    # Handle SPAWN actions
                    player._halite -= spawn_cost
                    board._add_ship(Ship(ShipId(create_uid()), shipyard.position, 0, player.id, board))
                # Clear the shipyard's action so it doesn't repeat the same action automatically
                shipyard.next_action = None

            for ship in player.ships:
                if ship.next_action == ShipAction.CONVERT:
                    # Can't convert on an existing shipyard but you can use halite in a ship to fund conversion
                    if ship.cell.shipyard_id is None and (ship.halite + player.halite) >= convert_cost:
                        # Handle CONVERT actions
                        delta_halite = ship.halite - convert_cost
                        # Excess halite leftover from conversion is added to the player's total only after all conversions have completed
                        # This is to prevent the edge case of chaining halite from one convert to fund other converts
                        leftover_convert_halite += max(delta_halite, 0)
                        player._halite += min(delta_halite, 0)
                        board._add_shipyard(Shipyard(ShipyardId(create_uid()), ship.position, player.id, board))
                        board._delete_ship(ship)
                elif ship.next_action is not None:
                    # If the action is not None and is not CONVERT it must be NORTH, SOUTH, EAST, or WEST
                    ship.cell._ship_id = None
                    ship._position = ship.position.translate(ship.next_action.to_point(), configuration.size)
                    ship._halite *= (1 - board.configuration.move_cost)
                    # We don't set the new cell's ship_id here as it would be overwritten by another ship in the case of collision.
                    # Later we'll iterate through all ships and re-set the cell._ship_id as appropriate.

            player._halite += leftover_convert_halite
            # Lets just check and make sure.
            assert player.halite >= 0

        def resolve_collision(ships: List[Ship]) -> Tuple[Optional[Ship], List[Ship]]:
            """
            Accepts the list of ships at a particular position (must not be empty).
            Returns the ship with the least halite or None in the case of a tie along with all other ships.
            """
            if len(ships) == 1:
                return ships[0], []
            ships_by_halite = group_by(ships, lambda ship: ship.halite)
            smallest_halite = min(ships_by_halite.keys())
            smallest_ships = ships_by_halite[smallest_halite]
            if len(smallest_ships) == 1:
                # There was a winner, return it
                winner = smallest_ships[0]
                return winner, [ship for ship in ships if ship != winner]
            # There was a tie for least halite, all are deleted
            return None, ships

        # Check for ship to ship collisions
        ship_collision_groups = group_by(board.ships.values(), lambda ship: ship.position)
        for position, collided_ships in ship_collision_groups.items():
            winner, deleted = resolve_collision(collided_ships)
            if winner is not None:
                winner.cell._ship_id = winner.id
            for ship in deleted:
                board._delete_ship(ship)
                if winner is not None:
                    # Winner takes deleted ships' halite
                    winner._halite += ship.halite

        # Check for ship to shipyard collisions
        for shipyard in list(board.shipyards.values()):
            ship = shipyard.cell.ship
            if ship is not None and ship.player_id != shipyard.player_id:
                # Ship to shipyard collision
                board._delete_shipyard(shipyard)
                board._delete_ship(ship)

        # Deposit halite from ships into shipyards
        for shipyard in list(board.shipyards.values()):
            ship = shipyard.cell.ship
            if ship is not None and ship.player_id == shipyard.player_id:
                shipyard.player._halite += ship.halite
                ship._halite = 0

        # Collect halite from cells into ships
        for ship in board.ships.values():
            cell = ship.cell
            delta_halite = int(cell.halite * configuration.collect_rate)
            if ship.next_action not in ShipAction.moves() and cell.shipyard_id is None and delta_halite > 0:
                ship._halite += delta_halite
                cell._halite -= delta_halite
            # Clear the ship's action so it doesn't repeat the same action automatically
            ship.next_action = None

        # Regenerate halite in cells
        for cell in board.cells.values():
            if cell.ship_id is None:
                next_halite = round(cell.halite * (1 + configuration.regen_rate), 3)
                cell._halite = min(next_halite, configuration.max_cell_halite)
                # Lets just check and make sure.
            assert cell.halite >= 0

        board._step += 1

        return board


def board_agent(agent: Callable[[Board], None]):
    """
    Decorator used to create an agent that modifies a board rather than an observation and a configuration
    Automatically returns the modified board's next actions

    @board_agent
    def my_agent(board: Board) -> None:
        ...
    """
    @wraps(agent)
    def agent_wrapper(obs, config) -> Dict[str, str]:
        board = Board(obs, config)
        agent(board)
        return board.current_player.next_actions

    if agent.__module__ is not None and agent.__module__ in sys.modules:
        setattr(sys.modules[agent.__module__], agent.__name__, agent_wrapper)
    return agent_wrapper
